dnl
dnl PAC_APPEND_UNIQ_TOKEN( VARIABLE_NAME, TOKEN )
dnl token is defined as string without any blank space in it.
dnl
AC_DEFUN(PAC_APPEND_UNIQ_TOKEN,[
dnl Since PAC_APPEND_UNIQ_TOKEN may be called by PAC_MAKE_UNIQ_STRING
dnl saved IFS should be in a different variable than MAKE_UNIQ_STRING's.
dnl Otherwise bad thing will happen.
save_IFS="$IFS"
dnl eval flags=\$$1
flags="[$]$1"
token=$2
dnl
IFS=' '
match=no
for flag in $flags ; do
    if test "$flag" = "$token" ; then
        match=yes
        break
    fi
done
if test "$match" = "no" ; then
    flags="$flags $token"
fi
dnl
dnl eval $1='$flags'
$1="$flags"
dnl
IFS="$save_IFS"
])dnl
dnl
dnl
dnl PAC_MAKE_UNIQ_STRING - Eliminate the duplicated substrings in the input
dnl                        string variable.
dnl Assumptions:
dnl 1) substring separator of the input string is blank space.
dnl 2) the input string does not contain any '\n' or '\&' character.
dnl
dnl The function has the following properties:
dnl 1) preserves the blank space within pair of double and single quotes. 
dnl 2) preserves the order of the unique substring.
dnl 3) elminated duplicated substring will be later in order.
dnl
dnl PAC_MAKE_UNIQ_STRING( VARIABLE_NAME, [REMOVED_NAME] )
dnl
dnl Input/Output
dnl VARIABLE_NAME - name of the variable containing duplicated substrings.
dnl                 On output, VARIABLE will contain no duplicated substring.
dnl REMOVED_NAME  - name of the (optional) variable containing the elminated
dnl                 duplicated substings.
dnl
AC_DEFUN(PAC_MAKE_UNIQ_STRING, [
saveIFS="$IFS"
dnl eval inflags=\$$1
inflags="[$]$1"
dnl Define separator to be used as replacement of blanks in "..." or '...'
pac_ifs='\&'
dnl
dnl For Double Quote: Replace "pppp  qqqq" as "pppp&&qqqq"
dnl
IFS=$pac_ifs
dnl Prepend '\&' before beginning '\"', Append '\&' after ending '\"'
outflags=`echo $inflags | sed -e 's%\"\([^\"]*\)\"%'"$pac_ifs"'&'"$pac_ifs"'%g'`
dnl
dnl Use \& as separator to locate "...." where ' ' will be replaced by '\&'
dnl
tokens=""
for outflag in $outflags ; do
    case $outflag in
    \"*\")
        dnl Replace ' ' by $pac_ifs
        flag=`echo $outflag | sed -e 's/ /'"$pac_ifs"'/g'`
        tokens=$tokens$flag
        ;;
    *)
        tokens=$tokens$outflag
        ;;
    esac
done
inflags=$tokens
dnl
dnl For Single Quote: Replace 'pppp  qqqq' as 'pppp&&qqqq'
dnl
IFS=$pac_ifs
dnl Prepend '\&' before beginning '\'', Append '\&' after ending '\''
outflags=`echo $inflags | sed -e "s%\'\([^\']*\)\'%""$pac_ifs""&""$pac_ifs""%g"`
dnl
dnl Use \& as separator to locate '....' where ' ' will be replaced by '\&'
dnl
tokens=""
for outflag in $outflags ; do
    case $outflag in
    \'*\')
        dnl Replace ' ' by $pac_ifs
        flag=`echo $outflag | sed -e 's/ /'"$pac_ifs"'/g'`
        tokens=$tokens$flag
        ;;
    *)
        tokens=$tokens$outflag
        ;;
    esac
done
dnl
dnl Compute the unique token list and duplicated token list.
dnl
IFS=' '
uniqtokens=""
dupltokens=""
for token in $tokens ; do
    match=no
    for uniqtoken in $uniqtokens ; do
        if test "$uniqtoken" = "$token" ; then
            match=yes
            break
        fi
    done
    if test "$match" = "yes" ; then
        dupltokens="$dupltokens $token"
    else
        uniqtokens="$uniqtokens $token"
    fi
done
dnl Restore $pac_ifs to the original, i.e. ' '.
uniqflags=`echo $uniqtokens | sed -e 's%'"$pac_ifs"'% %g'`
duplflags=`echo $dupltokens | sed -e 's%'"$pac_ifs"'% %g'`
dnl eval $1='$uniqflags'
$1="$uniqflags"
ifelse([$2],[],:,[$2="$duplflags"])
dnl
IFS="$saveIFS"
])dnl



dnl Nesting safe macros for saving variables
dnl Usage: PAC_PUSH_FLAG(CFLAGS)
AC_DEFUN([PAC_PUSH_FLAG],[
	if test -z "${pac_save_$1_nesting}" ; then
	   pac_save_$1_nesting=0
	fi
	eval pac_save_$1_${pac_save_$1_nesting}='"$$1"'
	pac_save_$1_nesting=`expr ${pac_save_$1_nesting} + 1`
])

dnl Usage: PAC_POP_FLAG(CFLAGS)
AC_DEFUN([PAC_POP_FLAG],[
	pac_save_$1_nesting=`expr ${pac_save_$1_nesting} - 1`
	eval $1="\$pac_save_$1_${pac_save_$1_nesting}"
	eval pac_save_$1_${pac_save_$1_nesting}=""
])

dnl Usage: PAC_PUSH_ALL_FLAGS
AC_DEFUN([PAC_PUSH_ALL_FLAGS],[
	PAC_PUSH_FLAG(CFLAGS)
	PAC_PUSH_FLAG(CPPFLAGS)
	PAC_PUSH_FLAG(CXXFLAGS)
	PAC_PUSH_FLAG(FFLAGS)
	PAC_PUSH_FLAG(FCFLAGS)
	PAC_PUSH_FLAG(LDFLAGS)
	PAC_PUSH_FLAG(LIBS)
])

dnl Usage: PAC_POP_ALL_FLAGS
AC_DEFUN([PAC_POP_ALL_FLAGS],[
	PAC_POP_FLAG(CFLAGS)
	PAC_POP_FLAG(CPPFLAGS)
	PAC_POP_FLAG(CXXFLAGS)
	PAC_POP_FLAG(FFLAGS)
	PAC_POP_FLAG(FCFLAGS)
	PAC_POP_FLAG(LDFLAGS)
	PAC_POP_FLAG(LIBS)
])

dnl PAC_PREFIX_FLAG - Save flag with a prefix
dnl Usage: PAC_PREFIX_FLAG(PREFIX, FLAG)
AC_DEFUN([PAC_PREFIX_FLAG],[
	$1_$2=$$2
	export $1_$2
	AC_SUBST($1_$2)
])

dnl PAC_PREFIX_ALL_FLAGS - Save flags with a prefix
dnl Usage: PAC_PREFIX_ALL_FLAGS(PREFIX)
AC_DEFUN([PAC_PREFIX_ALL_FLAGS],[
	PAC_PREFIX_FLAG($1, CFLAGS)
	PAC_PREFIX_FLAG($1, CPPFLAGS)
	PAC_PREFIX_FLAG($1, CXXFLAGS)
	PAC_PREFIX_FLAG($1, FFLAGS)
	PAC_PREFIX_FLAG($1, FCFLAGS)
	PAC_PREFIX_FLAG($1, F90FLAGS)
	PAC_PREFIX_FLAG($1, LDFLAGS)
	PAC_PREFIX_FLAG($1, LIBS)
])

dnl Usage: PAC_APPEND_FLAG([-02], [CFLAGS])
dnl appends the given argument to the specified shell variable unless the
dnl argument is already present in the variable
AC_DEFUN([PAC_APPEND_FLAG],[
	AC_REQUIRE([AC_PROG_FGREP])
	AS_IF(
		[echo "$$2" | $FGREP -e '$1' >/dev/null 2>&1],
		[echo "$2(='$$2') contains '$1', not appending" >&AS_MESSAGE_LOG_FD],
		[echo "$2(='$$2') does not contain '$1', appending" >&AS_MESSAGE_LOG_FD
		$2="$$2 $1"]
	)
])


dnl Usage: PAC_APPEND_FLAG([-02], [CFLAGS])
dnl appends the given argument to the specified shell variable unless the
dnl argument is already present in the variable
AC_DEFUN([PAC_APPEND_FLAG],[
        AC_REQUIRE([AC_PROG_FGREP])
        AS_IF(
                [echo "$$2" | $FGREP -e '$1' >/dev/null 2>&1],
                [echo "$2(='$$2') contains '$1', not appending" >&AS_MESSAGE_LOG_FD],
                [echo "$2(='$$2') does not contain '$1', appending" >&AS_MESSAGE_LOG_FD
                $2="$$2 $1"]
        )
])

dnl Usage: PAC_PREPEND_FLAG([-lpthread], [LIBS])
dnl Prepends the given argument to the specified shell variable unless the
dnl argument is already present in the variable.
dnl
dnl This is typically used for LIBS and similar variables because libraries
dnl should be added in reverse order.
AC_DEFUN([PAC_PREPEND_FLAG],[
        AC_REQUIRE([AC_PROG_FGREP])
        AS_IF(
                [echo "$$2" | $FGREP -e '$1' >/dev/null 2>&1],
                [echo "$2(='$$2') contains '$1', not prepending" >&AS_MESSAGE_LOG_FD],
                [echo "$2(='$$2') does not contain '$1', prepending" >&AS_MESSAGE_LOG_FD
                $2="$1 $$2"]
        )
])


