/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"
#include "mpicomm.h"

/* -- Begin Profiling Symbol Block for routine MPI_Comm_split */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Comm_split = PMPI_Comm_split
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Comm_split  MPI_Comm_split
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Comm_split as PMPI_Comm_split
#endif
/* -- End Profiling Symbol Block */

/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Comm_split
#define MPI_Comm_split PMPI_Comm_split

#endif

#undef FUNCNAME
#define FUNCNAME MPI_Comm_split


#if defined(FINEGRAIN_MPI)
#if 0
#define SPLIT_TIMING_START do{                       \
                               if (root == fgrank) { \
                                 starttime = MPI_Wtime();} \
                             }while(0)

#define SPLIT_TIMING_END(a_)  do{                               \
                                    if (root == fgrank) {       \
                                        endtime = MPI_Wtime();          \
                                        printf(#a_);printf(" Time = %f second\n", endtime-starttime); } \
                                }while(0)

#define LEADER_TIMING_START do{                     \
                               if ((local_leader_gid == fgrank) && (rank == 1)) { \
                                 starttime = MPI_Wtime();} \
                             }while(0)

#define LEADER_TIMING_END(a_)  do{                               \
                                   if ((local_leader_gid == fgrank) && (rank == 1)) {              \
                                        endtime = MPI_Wtime();          \
                                        printf(#a_);printf(" Leader_Time = %f second\n", endtime-starttime); } \
                                }while(0)

#define LEADER_COMP_MAP_FLAG(a_) do{                                    \
                                   if ((local_leader_gid == fgrank) && (rank == 1)) {              \
                                       printf("==================================  COMP_MAP rep flag is : %d   ======\n", a_->settype_flag); } \
                                }while(0)

#else
#define SPLIT_TIMING_START
#define SPLIT_TIMING_END(a_)
#define LEADER_TIMING_START
#define LEADER_TIMING_END(a_)
#define LEADER_COMP_MAP_FLAG(a_)
#endif

#define SCATTER_SIZE 4
#define GATHER_SIZE 3

typedef struct GroupGather{
    int color;
    int key;
    int gid;
    short int LBI;
    short int LBI_count;
    int pid;
    int wid;
} GroupGather;


typedef struct GroupInfoLinkedList{
    int *PIDarray;
    int *gid_array;
    int groupsize;
    struct GroupInfoLinkedList *next;
} GroupInfoLinkedList;



static int qsortcolorComparison(const void* groupgather1_ptr, const void* groupgather2_ptr){
    int c1 = ((GroupGather *)(groupgather1_ptr))->color;
    int c2 = ((GroupGather *)(groupgather2_ptr))->color;
    if(c1 < c2) return (-1);
    else if(c1 == c2) return (0);
    else return (1);
}


static inline void MPIU_Sort_intcolor( GroupGather *groupgather, int size )
{
    qsort(groupgather, size, sizeof(GroupGather), qsortcolorComparison);
}

static int qsortkeyComparison(const void* groupgather1_ptr, const void* groupgather2_ptr){
    int k1 = ((GroupGather *)(groupgather1_ptr))->key;
    int k2 = ((GroupGather *)(groupgather2_ptr))->key;
    if(k1 < k2) return (-1);
    else if(k1 == k2) return (0);
    else return (1);
}

static inline void MPIU_Sort_intkey( GroupGather *groupgather, int groupsize )
{
    qsort(groupgather, groupsize, sizeof(GroupGather), qsortkeyComparison);
}




static inline void Insert_GroupInfoLinkedList(GroupInfoLinkedList** groupinfo_head,  GroupInfoLinkedList** groupinfo_tail,  GroupInfoLinkedList* groupinfo_cell)
{
    if(*groupinfo_head){
        assert(*groupinfo_tail);
        (*groupinfo_tail)->next = groupinfo_cell;
        *groupinfo_tail = groupinfo_cell;
    }else{
        *groupinfo_head = *groupinfo_tail = groupinfo_cell;
    }
}

static inline void Free_GroupInfoLinkedList(GroupInfoLinkedList** groupinfo_head,  GroupInfoLinkedList** groupinfo_tail)
{
    GroupInfoLinkedList * tmp = *groupinfo_head;
    while(tmp){
        GroupInfoLinkedList * tmpprev = tmp;
        MPIU_Free(tmp->PIDarray);
        MPIU_Free(tmp->gid_array);
        tmp = tmp->next;
        MPIU_Free(tmpprev);
    }
    *groupinfo_head = *groupinfo_tail = NULL;
}

static inline void print_GroupInfoLinkedList(GroupInfoLinkedList* groupinfo_head,  GroupInfoLinkedList* groupinfo_tail, int pid_size)
{
    GroupInfoLinkedList * tmp = groupinfo_head;
    int i, cellnum = 1;
    int groupsize;
    while(tmp){
        groupsize = tmp->groupsize;
        for(i=0; i<pid_size; i++){
            if(tmp->PIDarray[i] == -1) continue;
            printf("Group#:%d...groupsize=%d...pid=%d, local_leader_gid=%d\n", cellnum, groupsize, i, tmp->PIDarray[i]);
        }
        for(i=0; i<groupsize; i++){
            printf("Group#:%d...groupsize=%d...gid_array[%d]=%d\t", cellnum, groupsize, i, tmp->gid_array[i]);
        }
        printf("\n");
        cellnum++;
        tmp = tmp->next;
    }
}

static inline void print_gidkey(int *gid, int *key, int size ){
    int i;
    for(i=0; i<size; i++){
        printf("key=%d, gid=%d\n", key[i], gid[i]);    
    }
}

static inline void print_groupgather(GroupGather *groupgather, int size ){
    int i;
    for(i=0; i<size; i++){
        printf("color=%d, key=%d, gid=%d, LBI=%d, LBI_count=%d, pid=%d, wid=%d\n", groupgather[i].color, groupgather[i].key, groupgather[i].gid, groupgather[i].LBI, groupgather[i].LBI_count, groupgather[i].pid, groupgather[i].wid);    
    }
}
#endif


#if defined(USE_LID_LBI_CONTEXT) /* HK: */

typedef struct splittype {
    int color, key;
    int count_LBI;
} splittype;



static int MPIR_Find_local_leader_and_CID( int root, int fgrank, int size, int totprocs, int *num_local_leaders,
                                           MPID_Comm *local_comm_ptr,
                                           GroupGather *groupgather, splittype *recvFromAll,
                                           GroupInfoLinkedList **groupinfo_head_dbptr, GroupInfoLinkedList **groupinfo_tail_dbptr,
                                           int *scatterarray ) 
{
    static const char FCNAME[] = "MPIR_Find_local_leader_and_CID";
    int mpi_errno = MPI_SUCCESS;
    double starttime, endtime;
    int scatterarrayfills = 0, i, j;
    int colorstart = 0, curr_index, CID, validcolorstart;
    SPLIT_TIMING_START;
    for(i=0; i<totprocs; i++){
        groupgather[i].color = recvFromAll[i].color;
        groupgather[i].key = recvFromAll[i].key;
        groupgather[i].gid = i;
        MPIR_Separate_count_LBI(recvFromAll[i].count_LBI, &(groupgather[i].LBI), &(groupgather[i].LBI_count));
        RTWPmapFind(local_comm_ptr->co_shared_vars->rtw_map, i, &(groupgather[i].wid), &(groupgather[i].pid)); /* HK: TODO RTWCompMapBlockFind */
    }
    MPIU_Free(recvFromAll);        
    SPLIT_TIMING_END(RTWPmapFind);
    
    SPLIT_TIMING_START;
    MPIU_Sort_intcolor(groupgather, totprocs);
    SPLIT_TIMING_END(MPIU_Sort_intcolor);
    
    SPLIT_TIMING_START;
    colorstart = 0;
    while( (groupgather[colorstart].color == MPI_UNDEFINED) && (colorstart < totprocs) ){
        colorstart++;
    }
    curr_index = validcolorstart = colorstart;
    
    while(curr_index < totprocs) {
        /* Each group has its own PIDarray */
        assert (colorstart == curr_index);
        int groupstart = colorstart;
        int groupsize = 0;    
        int curr_color = groupgather[groupstart].color;
        short int maxLBI_count = -1;
        short int maxLBI = -1;    
        int LID = -1;
        int *PIDarray = (int*) MPIU_Malloc(sizeof(int) * size);
        /* --BEGIN ERROR HANDLING-- */
        if (PIDarray == NULL) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        for (i=0; i<size; i++)
            {
                PIDarray[i] = -1; /* Initializer */
            }
        i = groupstart;
        while((i < totprocs) && (groupgather[i].color == curr_color)){
            /* Picking the one with highest count of 1's (i.e., least times leader) */
            if((maxLBI_count == -1) || (maxLBI_count < groupgather[i].LBI_count)){ 
                maxLBI_count = groupgather[i].LBI_count;
                maxLBI = groupgather[i].LBI;
                LID = groupgather[i].wid;
            }
            /* Do we need to break ties based on gid? Not doing that currently. Filling in
               the local_leader gid (the one with smallest gid) for each pid of that
               particular group. A value of -1 in PIDarray should be skipped.
            */
            if( (PIDarray[groupgather[i].pid] == -1) || (PIDarray[groupgather[i].pid] > groupgather[i].gid) ){
                if(PIDarray[groupgather[i].pid] == -1){
                    (*num_local_leaders)++;
                }
                PIDarray[groupgather[i].pid] = groupgather[i].gid;
            }
            i++;
            groupsize++;
            curr_index++;
        }
        /* Construct CID */
        assert(LID >=0);
        if(maxLBI_count == 0){
            assert(maxLBI < 0);
            printf("Failed! Too many communicators! LBI_count is zero for all group members.\n");
            MPIU_ERR_CHKANDJUMP(maxLBI_count == 0, mpi_errno, MPI_ERR_OTHER, 
                                "**toomanycomm" );
            exit(-1); /* Just in case the above falls through */
        }else{
            assert( (maxLBI >= 0) && (maxLBI < (MPIR_CONTEXT_INT_BITS*MAX_LBI_MASK)) );
            assert(maxLBI_count > 0);
        }
        CID = MPIR_Get_CID(LID, maxLBI);
        
        int *gid_array = (int*) MPIU_Malloc(sizeof(int) * groupsize);
        /* --BEGIN ERROR HANDLING-- */
        if (gid_array == NULL) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        MPIU_Sort_intkey(groupgather+groupstart, groupsize);
        
        /* Fill in the scatter array */
        for(i=groupstart, j=0; i<(groupstart+groupsize); i++, j++){
            int local_leadergid = PIDarray[groupgather[i].pid];
            int gid = groupgather[i].gid;
            assert( (scatterarray[gid*SCATTER_SIZE]==-1) && (scatterarray[gid*SCATTER_SIZE+1]==-1)
                    && (scatterarray[gid*SCATTER_SIZE+2]==-1) && (scatterarray[gid*SCATTER_SIZE+3]==-1) );
            scatterarray[gid*SCATTER_SIZE] = CID;
            scatterarray[gid*SCATTER_SIZE + 1] = local_leadergid;
            scatterarray[gid*SCATTER_SIZE + 2] = groupsize;
            scatterarrayfills++;
            /* Fill in gid_array for MPIC_Isend */
            gid_array[j] = gid;
        }
        
        
        for(i=0; i<groupsize; i++){
            int outer_gid = gid_array[i];
            scatterarray[outer_gid*SCATTER_SIZE+3] = i; /* local rank in the new group */
        }
        
        GroupInfoLinkedList * groupinfo_cell = (GroupInfoLinkedList *) MPIU_Malloc(sizeof(GroupInfoLinkedList));
        /* --BEGIN ERROR HANDLING-- */
        if (groupinfo_cell == NULL) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        
        groupinfo_cell->PIDarray = PIDarray;
        groupinfo_cell->gid_array = gid_array;
        groupinfo_cell->groupsize = groupsize;
        groupinfo_cell->next = NULL;
        Insert_GroupInfoLinkedList(groupinfo_head_dbptr, groupinfo_tail_dbptr, groupinfo_cell);
        
        colorstart += groupsize;
    }    
    assert(scatterarrayfills == (totprocs-validcolorstart)); 
    MPIU_Free(groupgather);
    
    SPLIT_TIMING_END(Computation+MPIU_Sort_intkey);


    /* ... end of body of routine ... */
    
 fn_exit:
    return mpi_errno;
    
 fn_fail:
    printf("%s: my_fgrank=%d ...................... Arrived at fn_fail .....................\n", __FUNCTION__, my_fgrank);
    goto fn_exit;    
}



/*@

MPI_Comm_split - Creates new communicators based on colors and keys

Input Parameters:
+ comm - communicator (handle) 
. color - control of subset assignment (nonnegative integer).  Processes 
  with the same color are in the same new communicator 
- key - control of rank assigment (integer)

Output Parameter:
. newcomm - new communicator (handle) 

Notes:
  The 'color' must be non-negative or 'MPI_UNDEFINED'.

Algorithm:
.vb
   1. The root gathers all the group membership information (color, key) and the LBI index and LBI count
      Root computes the Cid and the local leaders per process for each new communicator.
   2. Root scatters the Cid, leader_gid, group size and the local_fgrank in new communicator to all.

   3. Root sends the leaders their group membership information.

   Now, since the Cid is globally unique, I will use the Cid to synchronize the non-leaders and the
   leader in the new communicator.
   The leader will create a shared map and store a pointer to it in the lookup hash using Cid as key.
   The non-leaders will use Cid as key for lookup hash. If an entry is found, they will know that
   the leader has created the shared map and they can go ahead and cache it locally. Otherwise they
   will simply yield until the time a valid entry is found in the lookup hash for that Cid.

   4. If the local proclet has a color of MPI_UNDEFINED, it can return a NULL comm. 

.ve
 
.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_free 

@*/

int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm) /* HK: */
{
    static const char FCNAME[] = "MPI_Comm_split";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL, *newcomm_ptr, *local_comm_ptr;
    MPI_Comm  local_comm;
    splittype *sendToRoot = NULL, *recvFromAll = NULL;
    int size, remote_size, i, j, k, new_remote_size;
    int fgrank = -1, totprocs = -1, leader_wid = -1;
    int root = 0;
    GroupGather *groupgather = NULL;
    int *scatterarray = NULL, *new_grouparray = NULL;
    int scatter_recv[SCATTER_SIZE] = {-3,-3,-3,-3};
    GroupInfoLinkedList *groupinfo_head = NULL, *groupinfo_tail = NULL;
    int num_local_leaders = 0;    
    int local_leader_gid = -1, new_groupsize = -1;
    int chosen_LID_for_CID = -1;
    short int chosen_LBI_for_CID = -1;
    double starttime, endtime;
    
    MPIU_THREADPRIV_DECL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPLIT);

    MPIR_ERRTEST_INITIALIZED_ORDIE();
    
    MPIU_THREAD_SINGLE_CS_ENTER("comm");
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPLIT);

    /* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    MPIR_ERRTEST_COMM(comm, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
    }
    
#   endif /* HAVE_ERROR_CHECKING */
    
    /* Get handles to MPI objects. */
    MPID_Comm_get_ptr( comm, comm_ptr );
    
    /* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
            /* Validate comm_ptr */
            MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
	    /* If comm_ptr is not valid, it will be reset to null */
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   endif /* HAVE_ERROR_CHECKING */

    /* ... body of routine ...  */
    
    size        = comm_ptr->local_size; 
    remote_size = comm_ptr->remote_size; /* HK: TODO. Used for intercomm */
    fgrank      = comm_ptr->rank;
    totprocs    = comm_ptr->totprocs;
    root        = 0;

    if(initialize_LBI_mask){
        MPIR_Init_LBI();
    }
    assert (LBI_mask != NULL);


    /* Root finds out the color, keys and LBI count and LBI index of all of the proclets. */
    sendToRoot = (splittype*) MPIU_Malloc(sizeof(splittype));
    /* --BEGIN ERROR HANDLING-- */
    if (sendToRoot == NULL) {
        printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
        mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
        goto fn_fail;
    }
    /* --END ERROR HANDLING-- */
    sendToRoot->color = color;
    sendToRoot->key   = key;
    sendToRoot->count_LBI = MPIR_Get_count_LBI_sndval(LBI_mask);
    
    if(root == fgrank){
        recvFromAll = (splittype*) MPIU_Malloc(totprocs*sizeof(splittype));
        groupgather = (GroupGather*) MPIU_Malloc(sizeof(GroupGather) * totprocs);
        scatterarray = (int*) MPIU_Calloc(SCATTER_SIZE * totprocs, sizeof(int));
        /* --BEGIN ERROR HANDLING-- */
        if ( (recvFromAll == NULL) || (groupgather == NULL) || (scatterarray == NULL) ) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        for (i=0; i<(SCATTER_SIZE * totprocs); i++){
            scatterarray[i] = -1; /* Initializer */
        }
    }

    
    MPIU_THREADPRIV_GET;

    /* Get the communicator to use in collectives on the local group of 
       processes */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */
	if (!comm_ptr->local_comm) {
	    MPIR_Setup_intercomm_localcomm( comm_ptr );
	}
	local_comm_ptr = comm_ptr->local_comm;
	local_comm     = local_comm_ptr->handle;
    }
    else {
	local_comm_ptr = comm_ptr;
	local_comm     = comm;
    }

    /* Root gathers information on the local group of processes */
    SPLIT_TIMING_START;
#if 1
    mpi_errno = MPI_Gather(sendToRoot, GATHER_SIZE, MPI_INT, recvFromAll, GATHER_SIZE, MPI_INT, root, local_comm );
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;

#else /* As an alternative to MPI_Gather use explicit send/receives as below */

    if (root == fgrank){                   
 #if 1
        /* Root uses MPIC_Irecv. */
        MPI_Request *recvrequests = MPIU_Malloc(sizeof(MPI_Request)*totprocs);
        MPI_Status *recvrequeststats = MPIU_Malloc(sizeof(MPI_Status)*totprocs);
        /* --BEGIN ERROR HANDLING-- */
        if ( (recvrequests == NULL) || (recvrequeststats == NULL) ) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        

        for(i=1, k=0; i<(totprocs); i++){ /* Note: Starting from i=1, root is not recveing from itself, copying below */ 
            mpi_errno = MPIC_Irecv(recvFromAll+i, GATHER_SIZE, MPI_INT, i, MPIR_COMM_SPLIT_TAG, local_comm, &recvrequests[k]);
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
            k++;
        }
        mpi_errno = MPI_Waitall(k, recvrequests,  recvrequeststats); /* HK: There is no MPIC_Waitall */
        if (mpi_errno != MPI_SUCCESS) goto fn_fail;
                
        if(recvrequests) MPIU_Free(recvrequests);
        if(recvrequeststats) MPIU_Free(recvrequeststats);                    
 #else /* Root uses MPIC_Recv */
        MPI_Status recvstatus;
        splittype tempRecvBuf;
        for(i=0; i<(totprocs-1); i++){
            /*
              HK: Note: Not using MPI_ANY_SOURCE in MPIC_Recv because that can lead to
              message mismatch and deadlock. Consider the following scenario:
              Proclet Px calls MPI_Comm_split() with color=MPI_UNDEFINED on local_comm
              and after sending its color info to root it leaves the function because
              it is not involved in any further communication.
              Px makes another call to MPI_Comm_split() immediately after that on the
              same local_comm (and hence the same root). It is possible that Px is
              scheduled again before the rest of proclets are done with previous
              MPI_Comm_split() operation. Using MPI_ANY_SOURCE can result in Px's
              second color info message being received instead of the correct
              proclets's one. 
            */
            MPIC_Recv( &tempRecvBuf, GATHER_SIZE, MPI_INT, i, MPIR_COMM_SPLIT_TAG, local_comm, &recvstatus );
            assert ( recvstatus.MPI_SOURCE == i );
            recvFromAll[i] = tempRecvBuf;
        }
 #endif
        recvFromAll[root] = *sendToRoot; /* Root copies its own color, key and LBI information */
    }
    else{
        MPIC_Send( sendToRoot, GATHER_SIZE, MPI_INT, root, MPIR_COMM_SPLIT_TAG, local_comm);
    }
    
#endif
    SPLIT_TIMING_END(Gather);

    if (root == fgrank) {

        assert ( (recvFromAll[root].color == color) && (recvFromAll[root].key == key)
                 && (recvFromAll[root].count_LBI == sendToRoot->count_LBI) );

        mpi_errno = MPIR_Find_local_leader_and_CID(root, fgrank, size, totprocs, &num_local_leaders,
                                                   local_comm_ptr,
                                                   groupgather, recvFromAll,
                                                   &groupinfo_head, &groupinfo_tail,
                                                   scatterarray );
        if (mpi_errno != MPI_SUCCESS) goto fn_fail;
    }        
    

    MPIU_Free(sendToRoot);    

    /* Root scatters the CID, local_leadergid and groupsize to all in the outer communicator. */
    SPLIT_TIMING_START;
#if 0
    mpi_errno = MPI_Scatter(scatterarray, SCATTER_SIZE, MPI_INT, scatter_recv, SCATTER_SIZE, MPI_INT, root, local_comm);
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;
    
#else
    
    if (root == fgrank) {
 #if 0 /* Root uses MPIC_Send. Performs worse than MPIC_Isend below. */
        for(i=1; i<(totprocs); i++){ /* Note: Starting from i=1, root is not sending to itself, copying below */
            if(scatterarray[i*SCATTER_SIZE] == -1){
                for(j=1; j<SCATTER_SIZE; j++){
                    assert(scatterarray[i*SCATTER_SIZE+j] == -1);
                }
                continue;
            }
            mpi_errno = MPIC_Send(scatterarray+(i*SCATTER_SIZE), SCATTER_SIZE, MPI_INT, i, MPIR_COMM_SPLIT_TAG, local_comm);
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
        }
 #else
        /* Root uses MPIC_Isend. */
        MPI_Request *sndrequests = MPIU_Malloc(sizeof(MPI_Request)*totprocs);
        MPI_Status *requeststats = MPIU_Malloc(sizeof(MPI_Status)*totprocs);
        /* --BEGIN ERROR HANDLING-- */
        if ( (sndrequests == NULL) || (requeststats == NULL) ) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        

        for(i=1, k=0; i<(totprocs); i++){ /* Note: Starting from i=1, root is not sending to itself, copying below */ 
            if(scatterarray[i*SCATTER_SIZE] == -1){
                for(j=1; j<SCATTER_SIZE; j++){
                    assert(scatterarray[i*SCATTER_SIZE+j] == -1);
                }
                continue;
            }
            mpi_errno = MPIC_Isend(scatterarray+(i*SCATTER_SIZE), SCATTER_SIZE, MPI_INT, i, MPIR_COMM_SPLIT_TAG, local_comm, &sndrequests[k]);
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
            k++;
        }

        mpi_errno = MPI_Waitall(k, sndrequests,  requeststats); /* HK: There is no MPIC_Waitall */
        if (mpi_errno != MPI_SUCCESS) goto fn_fail;

        
        if(sndrequests) MPIU_Free(sndrequests);
        if(requeststats) MPIU_Free(requeststats);
 #endif

        
        for(i=0; i<SCATTER_SIZE; i++){
            scatter_recv[i] = scatterarray[root*SCATTER_SIZE+i]; /* Root copies its own scatter information */
        }
    }
    else {
        if(color != MPI_UNDEFINED){
            MPI_Status recvstatus;
            MPIC_Recv( scatter_recv, SCATTER_SIZE, MPI_INT, root, MPIR_COMM_SPLIT_TAG, local_comm, &recvstatus );
        }
    }
    
#endif
    SPLIT_TIMING_END(Scatter);
    
    if(root == fgrank){
        MPIU_Free(scatterarray);   
    }
    
    
    FGP_DBG_PRINTF("%s: my_fgrank=%d. fgrank=%d, CID=scatter_recv[0]=%d, local_leader_gid=scatter_recv[1]=%d, new_groupsize=scatter_recv[2]=%d, local_fgrank=scatter_recv[3]=%d\n",
           __FUNCTION__, my_fgrank, fgrank, scatter_recv[0], scatter_recv[1], scatter_recv[2], scatter_recv[3]);

    if (color != MPI_UNDEFINED){
        assert( (scatter_recv[0] > 0) && (scatter_recv[2] > 0) );
        assert( (scatter_recv[1] >= 0) && (scatter_recv[3] >= 0) );
    }else{
        assert( (scatter_recv[0] < 0) && (scatter_recv[1] < 0) && (scatter_recv[2] < 0) && (scatter_recv[3] < 0) );
    }

    
    local_leader_gid = scatter_recv[1];
    new_groupsize = scatter_recv[2];
    
    
    /* If we're an intercomm, we need to do the same thing for the remote
       table, as we need to know the size of the remote group of the
       same color before deciding to create the communicator */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */

    }
    else {
	/* Set the size of the remote group to the size of our group.
	   This simplifies the test below for intercomms with an empty remote
	   group (must create comm_null) */
	new_remote_size = new_groupsize;
    }


    /* In the intercomm case, we need to exchange the context ids */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */

    }

    if (fgrank == root) {
        SPLIT_TIMING_START;
        MPI_Request *sndrequests = MPIU_Malloc(sizeof(MPI_Request)*num_local_leaders);
        MPI_Status *requeststats = MPIU_Malloc(sizeof(MPI_Status)*num_local_leaders);
        /* --BEGIN ERROR HANDLING-- */
        if ( (sndrequests == NULL) || (requeststats == NULL) ) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        GroupInfoLinkedList *tmp = groupinfo_head;
        int i, num_sends = 0, root_is_leader = 0;
        int curr_groupsize;
        while(tmp){
            curr_groupsize = tmp->groupsize;
            for(i=0; i<size; i++){
                if(tmp->PIDarray[i] == -1) continue;
                if (tmp->PIDarray[i] == root){ /* This would be true if color != MPI_UNDEFINED for root */
                     /* Root does not send to itself. Does a local copy */
                    assert (new_groupsize == curr_groupsize);
                    new_grouparray = MPIU_Malloc(sizeof(int)*curr_groupsize);
                    /* --BEGIN ERROR HANDLING-- */
                    if (new_grouparray == NULL) {
                        printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
                        mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                        goto fn_fail;
                    }
                    /* --END ERROR HANDLING-- */

                    memcpy(new_grouparray, tmp->gid_array, curr_groupsize*sizeof(int));
                    root_is_leader = 1;
                    continue;
                }
                mpi_errno = MPIC_Isend(tmp->gid_array, curr_groupsize, MPI_INT, tmp->PIDarray[i], MPIR_COMM_SPLIT_TAG, local_comm, &sndrequests[num_sends]);
                //mpi_errno = MPIC_Send(tmp->gid_array, curr_groupsize, MPI_INT, tmp->PIDarray[i], MPIR_COMM_SPLIT_TAG, local_comm);
                if (mpi_errno != MPI_SUCCESS) goto fn_fail;
                num_sends++;
            }
            tmp = tmp->next;
        }
        assert (root_is_leader == 0 || root_is_leader == 1);
        assert ( (num_sends + root_is_leader) == num_local_leaders);
        
        if (num_sends){
            mpi_errno = MPI_Waitall(num_sends, sndrequests,  requeststats); /* HK: There is no MPIC_Waitall */
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
        }
        

        if(sndrequests) MPIU_Free(sndrequests);
        if(requeststats) MPIU_Free(requeststats);
        
        SPLIT_TIMING_END(Sending_to_Leaders);

        Free_GroupInfoLinkedList(&groupinfo_head, &groupinfo_tail);
                
    }
    else if (fgrank == local_leader_gid){  /* HK: else-if here because the root is not receiving from itself */
        LEADER_TIMING_START;
        new_grouparray = MPIU_Malloc(sizeof(int)*new_groupsize);
        /* --BEGIN ERROR HANDLING-- */
        if (new_grouparray == NULL) {
            printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
            mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
            goto fn_fail;
        }
        /* --END ERROR HANDLING-- */
        
        MPI_Status status;
        /* HK: Note: Using MPIC_Recv below, since this is a recv inside a collective operation. */        
        mpi_errno = MPIC_Recv(new_grouparray, new_groupsize, MPI_INT, root, MPIR_COMM_SPLIT_TAG, local_comm, &status);
        if (mpi_errno != MPI_SUCCESS) goto fn_fail;
        LEADER_TIMING_END(Receive_Map);
    }    
        
    /* Now, create the new communicator structure if necessary */
    if (color != MPI_UNDEFINED && new_remote_size > 0) {
        
        mpi_errno = MPIR_Comm_create( &newcomm_ptr );
        if (mpi_errno) goto fn_fail;
        
        newcomm_ptr->context_id	    = scatter_recv[0]; /* CID */
        newcomm_ptr->local_size	    = size; /* HK: This is the total number of HWP processes */
        newcomm_ptr->comm_kind	    = comm_ptr->comm_kind;
        
        /* Other fields depend on whether this is an intercomm or intracomm */
        if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */
            
        }
	else {
	    /* INTRA Communicator */
            newcomm_ptr->recvcontext_id = newcomm_ptr->context_id;
            newcomm_ptr->remote_size = size; /* HK: TODO Double-check. This is the total number of HWP processes */
            newcomm_ptr->rank = scatter_recv[3]; 
            MPIR_Intra_newcomm_init(newcomm_ptr, comm_ptr->p_rank, new_groupsize); /* _p_rank_ */

            /* Update of LBI mask by the chosen CID leader */
            int chosen_LID_for_CID = -1;
            short int chosen_LBI_for_CID = -1;
            MPIR_Separate_CID_into_LID_LBI(newcomm_ptr->context_id, &chosen_LID_for_CID, &chosen_LBI_for_CID);
            assert( (chosen_LID_for_CID >=0) && (chosen_LBI_for_CID >=0) &&
                    (chosen_LBI_for_CID < (MPIR_CONTEXT_INT_BITS*MAX_LBI_MASK)) );             
            /* Remember chosen_LID_for_CID is a worldrank, so comparing against my_fgrank */
            if (my_fgrank == chosen_LID_for_CID) {
                /* unset the bit location in chosen_LBI_for_CID */
                MPIR_Unsetbit_in_LBI_mask(LBI_mask, chosen_LBI_for_CID);
            }
            
            /* HK: Looking up pointer to shared rtw map in cidLookup_hshtbl.
               If it does not exist (i.e. first proclet in MPI_Comm_split()), then create it. */
            cidLookupHashItemptr stored = NULL;
            CidLookupHashFind(cidLookup_hshtbl, newcomm_ptr->context_id, &stored);
            if (fgrank != local_leader_gid) {
                while (stored == NULL){
                    FG_Yield();
                    CidLookupHashFind(cidLookup_hshtbl, newcomm_ptr->context_id, &stored);
                }
                MPIR_Comm_add_coshared_all_ref(stored->coproclet_shared_vars);
                newcomm_ptr->co_shared_vars = ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars));
                MPIR_Copy_ch3i_comm_ch(newcomm_ptr, ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars))->ch3i_ptr);
                FGP_DBG_PRINTF("Already present in cidLookup_hshtbl, context-id=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, newcomm_ptr->co_shared_vars->rtw_map);
            }
            else {
                LEADER_TIMING_START;
                assert(stored == NULL);
                
                newcomm_ptr->co_shared_vars = (Coproclet_shared_vars_t *)MPIU_Malloc(sizeof(Coproclet_shared_vars_t));
                /* --BEGIN ERROR HANDLING-- */
                if ((newcomm_ptr->co_shared_vars == NULL)) {                    
                    printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
                    mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                    goto fn_fail;
                }
                /* --END ERROR HANDLING-- */

                newcomm_ptr->co_shared_vars->rtw_map = (RTWmap *)RTWmapCreate(new_groupsize);
                assert( newcomm_ptr->co_shared_vars->rtw_map != NULL );
                int *rtw_blockinsert = (int*) MPIU_Malloc(sizeof(int) * new_groupsize);
                assert(rtw_blockinsert != 0);
                for (i=0; i<newcomm_ptr->totprocs; i++) {
                    /* Initializing worldrank to -1 */
                    int worldrank = -1, key;
                    key = new_grouparray[i];
                    RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, key, &worldrank, NULL);
                    key = i; /* HK: This is the local fgrank relative to this new group */
                    /* HK: Store item in newcomm_ptr rtw map which has i as
                       the local fgrank relative to the new communicator.
                       The corresponding worldrank in outer communicator is
                       determined and saved against fgrank=i in the newcomm rtw map. 
                    */
                    rtw_blockinsert[key] = worldrank;                    
                }
                
                RTWmapBlockInsert(newcomm_ptr->co_shared_vars->rtw_map, new_groupsize, rtw_blockinsert);
                MPIU_Free(rtw_blockinsert);
                
                newcomm_ptr->co_shared_vars->co_barrier_vars =  (struct coproclet_barrier_vars *)MPIU_Malloc(sizeof(struct coproclet_barrier_vars));

                /* --BEGIN ERROR HANDLING-- */
                if ((newcomm_ptr->co_shared_vars->co_barrier_vars == NULL)) {                    
                    printf("%s:Line_%d my_fgrank=%d, fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank, fgrank);
                    mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                    goto fn_fail;
                }
                /* --END ERROR HANDLING-- */
                
                newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal = 0;
                newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter = 0;
                newcomm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 0;
                /* Notify the device of this new communicator */
                MPID_Dev_comm_create_hook( newcomm_ptr );

                MPIR_Comm_init_coshared_all_ref(newcomm_ptr->co_shared_vars);
                newcomm_ptr->co_shared_vars->ch3i_ptr = &(newcomm_ptr->ch);
                stored = NULL;
                CidLookupHashInsert(cidLookup_hshtbl, newcomm_ptr->context_id, newcomm_ptr->co_shared_vars, &stored);
                if(!stored){
                    printf("Store failure in contextLeader_hshtbl\n");
                    MPIU_ERR_CHKANDJUMP(stored == 0, mpi_errno, MPI_ERR_OTHER, 
                                        "**hshstorefail" );
                    exit(-1);
                }
                FGP_DBG_PRINTF("Inserting in cidLookup_hshtbl, context-id=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, newcomm_ptr->co_shared_vars->rtw_map);
                
                LEADER_TIMING_END(RTWmapCreateBlockInsert);
#if defined(COMP_MAP)
                LEADER_COMP_MAP_FLAG(newcomm_ptr->co_shared_vars->rtw_map);
#endif
            }

	}
        

	/* Inherit the error handler (if any) */
	newcomm_ptr->errhandler = comm_ptr->errhandler;
	if (comm_ptr->errhandler) {
	    MPIR_Errhandler_add_ref( comm_ptr->errhandler );
	}
	
	*newcomm = newcomm_ptr->handle;
    }
    else {
	/* color was MPI_UNDEFINED.*/
	*newcomm = MPI_COMM_NULL;
        FGP_DBG_PRINTF("%s: newcomm is MPI_COMM_NULL ...........color=%d, new_remote_size=%d\n", __FUNCTION__, color, new_remote_size); 
    }
    
    if(new_grouparray){
        MPIU_Free(new_grouparray);
    }

    /* ... end of body of routine ... */

  fn_exit:
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPLIT);
    MPIU_THREAD_SINGLE_CS_EXIT("comm");
    return mpi_errno;
    
  fn_fail:
    printf("%s: my_fgrank=%d ...................... Arrived at fn_fail .....................\n", __FUNCTION__, my_fgrank);
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
	    "**mpi_comm_split",
	    "**mpi_comm_split %C %d %d %p", comm, color, key, newcomm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}



#else /* HK: #else USE_LID_LBI_CONTEXT is not defined, use MPICH2's context_id generation algorithm. */

typedef struct splittype {
    int color, key;
} splittype;

PMPI_LOCAL void MPIU_Sort_inttable( splittype *, int );
#ifndef MPICH_MPI_FROM_PMPI
/* Sort the entries in keytable into increasing order by key.  A stable
   sort should be used incase the key values are not unique. */
PMPI_LOCAL void MPIU_Sort_inttable( splittype *keytable, int size )
{
    splittype tmp;
    int i, j;

    /* FIXME Bubble sort */
    for (i=0; i<size; i++) {
	for (j=i+1; j<size; j++) {
	    if (keytable[i].key > keytable[j].key) {
		tmp	    = keytable[i];
		keytable[i] = keytable[j];
		keytable[j] = tmp;
	    }
	}
    }
}
#endif


/*@

MPI_Comm_split - Creates new communicators based on colors and keys

Input Parameters:
+ comm - communicator (handle) 
. color - control of subset assignment (nonnegative integer).  Processes 
  with the same color are in the same new communicator 
- key - control of rank assigment (integer)

Output Parameter:
. newcomm - new communicator (handle) 

Notes:
  The 'color' must be non-negative or 'MPI_UNDEFINED'.

.N ThreadSafe

.N Fortran

Algorithm:
.vb
  1. Use MPI_Allgather to get the color and key from each process
  2. Count the number of processes with the same color; create a 
     communicator with that many processes.  If this process has
     'MPI_UNDEFINED' as the color, create a process with a single member.
  3. Use key to order the ranks
  4. Set the VCRs using the ordered key values
.ve
 
.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Comm_free
@*/
int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm)
{
    static const char FCNAME[] = "MPI_Comm_split";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL, *newcomm_ptr, *local_comm_ptr;
    MPI_Comm  local_comm;
    splittype *table, *remotetable=0, *keytable, *remotekeytable=0;
    int       rank, size, remote_size, i, new_size, new_remote_size, 
	first_entry = 0, first_remote_entry = 0,
	*last_ptr;
    int       new_context_id, remote_context_id;
#if defined(FINEGRAIN_MPI)
    int totprocs = -1, leader_wid = -1;
#endif
    MPIU_THREADPRIV_DECL;
    MPIU_CHKLMEM_DECL(4);
    MPID_MPI_STATE_DECL(MPID_STATE_MPI_COMM_SPLIT);

    MPIR_ERRTEST_INITIALIZED_ORDIE();
    
    MPIU_THREAD_SINGLE_CS_ENTER("comm");
    MPID_MPI_FUNC_ENTER(MPID_STATE_MPI_COMM_SPLIT);

    /* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    MPIR_ERRTEST_COMM(comm, mpi_errno);
            if (mpi_errno) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
    }
    
#   endif /* HAVE_ERROR_CHECKING */
    
    /* Get handles to MPI objects. */
    MPID_Comm_get_ptr( comm, comm_ptr );
    
    /* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
            /* Validate comm_ptr */
            MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
	    /* If comm_ptr is not valid, it will be reset to null */
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   endif /* HAVE_ERROR_CHECKING */

    /* ... body of routine ...  */
    
    rank        = comm_ptr->rank;
    size        = comm_ptr->local_size; 
    remote_size = comm_ptr->remote_size; /* HK: TODO. Used for intercomm */

#if defined(FINEGRAIN_MPI)
    totprocs = comm_ptr->totprocs;
    /* Step 1: Find out what color and keys all of the processes have */
    MPIU_CHKLMEM_MALLOC(table,splittype*,totprocs*sizeof(splittype),mpi_errno,
			"table");
#else    
    /* Step 1: Find out what color and keys all of the processes have */
    MPIU_CHKLMEM_MALLOC(table,splittype*,size*sizeof(splittype),mpi_errno,
			"table");
#endif
    
    table[rank].color = color;
    table[rank].key   = key;

    MPIU_THREADPRIV_GET;

    /* Get the communicator to use in collectives on the local group of 
       processes */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */
	if (!comm_ptr->local_comm) {
	    MPIR_Setup_intercomm_localcomm( comm_ptr );
	}
	local_comm_ptr = comm_ptr->local_comm;
	local_comm     = local_comm_ptr->handle;
    }
    else {
	local_comm_ptr = comm_ptr;
	local_comm     = comm;
    }
    /* Gather information on the local group of processes */
    MPIR_Nest_incr();
    NMPI_Allgather( MPI_IN_PLACE, 2, MPI_INT, table, 2, MPI_INT, local_comm );
    MPIR_Nest_decr();

    /* Step 2: How many processes have our same color? */
    new_size = 0;
    if (color != MPI_UNDEFINED) {
	/* Also replace the color value with the index of the *next* value
	   in this set.  The integer first_entry is the index of the 
	   first element */
	last_ptr = &first_entry;
#if defined(FINEGRAIN_MPI)
        for (i=0; i<totprocs; i++) 
#else
	for (i=0; i<size; i++) 
#endif
        {
	    /* Replace color with the index in table of the next item
	       of the same color.  We use this to efficiently populate 
	       the keyval table */
	    if (table[i].color == color) {
		new_size++;
		*last_ptr = i;
		last_ptr  = &table[i].color;
#if defined(FINEGRAIN_MPI)
                int worldrank = -1, key;
                key = i;                
                RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, key, &worldrank, NULL);
                if( (leader_wid == -1) || (leader_wid > worldrank) ){
                    leader_wid = worldrank;
                }
#endif
	    }
	}
#if defined(FINEGRAIN_MPI)
        assert(0 != new_size);
#endif
    }
    
    /* We don't need to set the last value to -1 because we loop through
       the list for only the known size of the group */

    /* If we're an intercomm, we need to do the same thing for the remote
       table, as we need to know the size of the remote group of the
       same color before deciding to create the communicator */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */
	splittype mypair;
	/* For the remote group, the situation is more complicated.
	   We need to find the size of our "partner" group in the
	   remote comm.  The easiest way (in terms of code) is for
	   every process to essentially repeat the operation for the
	   local group - perform an (intercommunicator) all gather
	   of the color and rank information for the remote group.
	*/
	MPIU_CHKLMEM_MALLOC(remotetable,splittype*,
			    remote_size*sizeof(splittype),mpi_errno,
			    "remotetable");
	MPIR_Nest_incr();
	/* This is an intercommunicator allgather */
	
	/* We must use a local splittype because we've already modified the
	   entries in table to indicate the location of the next rank of the
	   same color */
	mypair.color = color;
	mypair.key   = key;
	NMPI_Allgather( &mypair, 2, MPI_INT, remotetable, 2, MPI_INT,
			comm );
	MPIR_Nest_decr();

	/* Each process can now match its color with the entries in the table */
	new_remote_size = 0;
	last_ptr = &first_remote_entry;
	for (i=0; i<remote_size; i++) {
	    /* Replace color with the index in table of the next item
	       of the same color.  We use this to efficiently populate 
	       the keyval table */
	    if (remotetable[i].color == color) {
		new_remote_size++;
		*last_ptr = i;
		last_ptr  = &remotetable[i].color;
	    }
	}
	/* Note that it might find that there a now processes in the remote
	   group with the same color.  In that case, COMM_SPLIT will
	   return a null communicator */
    }
    else {
	/* Set the size of the remote group to the size of our group.
	   This simplifies the test below for intercomms with an empty remote
	   group (must create comm_null) */
	new_remote_size = new_size;
    }

    /* Step 3: Create the communicator */
    /* Collectively create a new context id.  The same context id will
       be used by each (disjoint) collections of processes.  The
       processes whose color is MPI_UNDEFINED will return the 
       context id to the pool */
    /* In the multi-threaded case, MPIR_Get_contextid assumes that the
       calling routine already holds the single criticial section */
    new_context_id = MPIR_Get_contextid( local_comm_ptr );
    MPIU_ERR_CHKANDJUMP(new_context_id == 0, mpi_errno, MPI_ERR_OTHER, 
			"**toomanycomm" );

    /* In the intercomm case, we need to exchange the context ids */
    if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */
	if (comm_ptr->rank == 0) {
	    mpi_errno = MPIC_Sendrecv( &new_context_id, 1, MPI_INT, 0, 0,
				       &remote_context_id, 1, MPI_INT, 
				       0, 0, comm, MPI_STATUS_IGNORE );
	    if (mpi_errno) { MPIU_ERR_POP( mpi_errno ); }
            MPIR_Nest_incr();
	    NMPI_Bcast( &remote_context_id, 1, MPI_INT, 0, local_comm );
            MPIR_Nest_decr();
	}
	else {
	    /* Broadcast to the other members of the local group */
            MPIR_Nest_incr();
	    NMPI_Bcast( &remote_context_id, 1, MPI_INT, 0, local_comm );
            MPIR_Nest_decr();
	}
    }

    /* Now, create the new communicator structure if necessary */
    if (color != MPI_UNDEFINED && new_remote_size > 0) {
    
	mpi_errno = MPIR_Comm_create( &newcomm_ptr );
	if (mpi_errno) goto fn_fail;

	newcomm_ptr->context_id	    = new_context_id;
#if defined(FINEGRAIN_MPI)
        newcomm_ptr->local_size	    = size; /* HK: This is the total number of HWP processes */
#else
	newcomm_ptr->local_size	    = new_size;
#endif
	newcomm_ptr->comm_kind	    = comm_ptr->comm_kind;
	/* Other fields depend on whether this is an intercomm or intracomm */

	/* Step 4: Order the processes by their key values.  Sort the
	   list that is stored in table.  To simplify the sort, we 
	   extract the table into a smaller array and sort that.
	   Also, store in the "color" entry the rank in the input communicator
	   of the entry. */
	MPIU_CHKLMEM_MALLOC(keytable,splittype*,new_size*sizeof(splittype),
			    mpi_errno,"keytable");
	for (i=0; i<new_size; i++) {
	    keytable[i].key   = table[first_entry].key;
	    keytable[i].color = first_entry;
	    first_entry	      = table[first_entry].color;
	}

	/* sort key table.  The "color" entry is the rank of the corresponding
	   process in the input communicator */
	MPIU_Sort_inttable( keytable, new_size );

	if (comm_ptr->comm_kind == MPID_INTERCOMM) { /* HK: TODO IMPORTANT MPID_INTERCOMM */
	    MPIU_CHKLMEM_MALLOC(remotekeytable,splittype*,
				new_remote_size*sizeof(splittype),
				mpi_errno,"remote keytable");
	    for (i=0; i<new_remote_size; i++) {
		remotekeytable[i].key   = remotetable[first_remote_entry].key;
		remotekeytable[i].color = first_remote_entry;
		first_remote_entry	= remotetable[first_remote_entry].color;
	    }

	    /* sort key table.  The "color" entry is the rank of the
	       corresponding process in the input communicator */
	    MPIU_Sort_inttable( remotekeytable, new_remote_size );

	    MPID_VCRT_Create( new_size, &newcomm_ptr->local_vcrt );
	    MPID_VCRT_Get_ptr( newcomm_ptr->local_vcrt, 
			       &newcomm_ptr->local_vcr );
	    for (i=0; i<new_size; i++) {
		MPID_VCR_Dup( comm_ptr->local_vcr[keytable[i].color], 
			      &newcomm_ptr->local_vcr[i] );
		if (keytable[i].color == comm_ptr->rank) {
		    newcomm_ptr->rank = i;
		}
	    }
	    
	    /* For the remote group, the situation is more complicated.
	       We need to find the size of our "partner" group in the
	       remote comm.  The easiest way (in terms of code) is for
	       every process to essentially repeat the operation for the
	       local group - perform an (intercommunicator) all gather
	       of the color and rank information for the remote group.
	     */
	    /* We apply the same sorting algorithm to the entries that we've
	       found to get the correct order of the entries.

	       Note that if new_remote_size is 0 (no matching processes with
	       the same color in the remote group), then MPI_COMM_SPLIT
	       is required to return MPI_COMM_NULL instead of an intercomm 
	       with an empty remote group. */

	    MPID_VCRT_Create( new_remote_size, &newcomm_ptr->vcrt );
	    MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
	    for (i=0; i<new_remote_size; i++) {
		MPID_VCR_Dup( comm_ptr->vcr[remotekeytable[i].color], 
			      &newcomm_ptr->vcr[i] );
	    }

	    newcomm_ptr->recvcontext_id = remote_context_id;
	    newcomm_ptr->remote_size    = new_remote_size;
	    newcomm_ptr->local_comm     = 0;
	    newcomm_ptr->is_low_group   = comm_ptr->is_low_group;

	}
	else {
	    /* INTRA Communicator */
	    newcomm_ptr->recvcontext_id = new_context_id;
#if defined(FINEGRAIN_MPI)
            MPIR_Intra_newcomm_init(newcomm_ptr, comm_ptr->p_rank, new_size); /* _p_rank_ */
            /* HK: Replacing the following with pointers to vcrt_world and vcr_world
               in MPIR_Intra_newcomm_init() above.
               
               #if 0   
                 MPID_VCRT_Create( size, &newcomm_ptr->vcrt );
                 MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
                 for (i=0; i<size; i++) {
                   MPID_VCR_Dup( comm_ptr->vcr[i], &newcomm_ptr->vcr[i] );
                 }
               #endif 
            */
            newcomm_ptr->remote_size    = size; /* HK: TODO Double-check. This is the total number of HWP processes */
            for (i=0; i<new_size; i++) {
                if(keytable[i].color == comm_ptr->rank){
                    newcomm_ptr->rank = i;
                }
            }
#else
	    newcomm_ptr->remote_size    = new_size;
	    MPID_VCRT_Create( new_size, &newcomm_ptr->vcrt );
	    MPID_VCRT_Get_ptr( newcomm_ptr->vcrt, &newcomm_ptr->vcr );
	    for (i=0; i<new_size; i++) {
		MPID_VCR_Dup( comm_ptr->vcr[keytable[i].color], 
			      &newcomm_ptr->vcr[i] );
		if (keytable[i].color == comm_ptr->rank) {
		    newcomm_ptr->rank = i;
		}
	    }
#endif
	}

	/* Inherit the error handler (if any) */
	newcomm_ptr->errhandler = comm_ptr->errhandler;
	if (comm_ptr->errhandler) {
	    MPIR_Errhandler_add_ref( comm_ptr->errhandler );
	}


#if defined(FINEGRAIN_MPI)
            /* HK: Looking up pointer to shared rtw map in contextLeader_hshtbl.
               If it does not exist (i.e. first proclet in MPI_Comm_split()), then create it. */
            cLitemptr stored = NULL;
            CL_LookupHashFind(contextLeader_hshtbl, newcomm_ptr->context_id, leader_wid, &stored);
            if(stored){
                MPIR_Comm_add_coshared_all_ref(stored->coproclet_shared_vars);
                newcomm_ptr->co_shared_vars = ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars));
                MPIR_Copy_ch3i_comm_ch(newcomm_ptr, ((Coproclet_shared_vars_t *)(stored->coproclet_shared_vars))->ch3i_ptr);
                FGP_DBG_PRINTF("Already present in contextLeader_hshtbl, context-id=%d, LeaderWorldRank=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, LeaderWorldRank, newcomm_ptr->co_shared_vars->rtw_map);
            }
            else {
                newcomm_ptr->co_shared_vars = (Coproclet_shared_vars_t *)MPIU_Malloc(sizeof(Coproclet_shared_vars_t));
                /* --BEGIN ERROR HANDLING-- */
                if ( (newcomm_ptr->co_shared_vars == NULL) ) {                    
                    printf("%s:Line_%d my_fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank); 
                    mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                    goto fn_fail;
                }
                /* --END ERROR HANDLING-- */

                newcomm_ptr->co_shared_vars->rtw_map = (RTWmap *)RTWmapCreate(new_size);
                assert( newcomm_ptr->co_shared_vars->rtw_map != NULL );
                int *rtw_blockinsert = (int*) MPIU_Malloc(sizeof(int) * new_size);
                assert(rtw_blockinsert != 0);
                for (i=0; i<new_size; i++) {
                    /* Initializing worldrank to -1 */
                    int worldrank = -1, key;
                    key = keytable[i].color;
                    RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, key, &worldrank, NULL);
                    key = i; /* HK: This is the local fgrank relative to this new group */
                    /* HK: Store item in newcomm_ptr rtw map which has i as the local fgrank relative
                       to the new communicator (Note keytable[i].color was the local fgrank in the outer communicator.)
                       The corresponding worldrank in outer communicator is determined and saved
                       against fgrank=i in the newcomm rtw map. 

                       Replacing the following with block insert.
                       RTWmapInsert(newcomm_ptr->co_shared_vars->rtw_map, key, worldrank);
                    */
                    rtw_blockinsert[key] = worldrank;                    
                }

                RTWmapBlockInsert(newcomm_ptr->co_shared_vars->rtw_map, new_size, rtw_blockinsert);
                MPIU_Free(rtw_blockinsert);                
                
                newcomm_ptr->co_shared_vars->co_barrier_vars =  (struct coproclet_barrier_vars *)MPIU_Malloc(sizeof(struct coproclet_barrier_vars));
                /* --BEGIN ERROR HANDLING-- */
                if ( (newcomm_ptr->co_shared_vars == NULL) ) {                    
                    printf("%s:Line_%d my_fgrank=%d, Failed! No memory!\n", __FUNCTION__, __LINE__, my_fgrank); 
                    mpi_errno = MPIR_Err_create_code( MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**nomem", 0 );
                    goto fn_fail;
                }
                /* --END ERROR HANDLING-- */

                newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal = 0;
                newcomm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter = 0;
                newcomm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 0;
                /* Notify the device of this new communicator */
                MPID_Dev_comm_create_hook( newcomm_ptr );

                MPIR_Comm_init_coshared_all_ref(newcomm_ptr->co_shared_vars);
                newcomm_ptr->co_shared_vars->ch3i_ptr = &(newcomm_ptr->ch);
                stored = NULL;
                CL_LookupHashInsert(contextLeader_hshtbl, newcomm_ptr->context_id, leader_wid, newcomm_ptr->co_shared_vars, &stored);
                if(!stored){
                    printf("Store failure in contextLeader_hshtbl\n");
                    MPIU_ERR_CHKANDJUMP(stored == 0, mpi_errno, MPI_ERR_OTHER, 
                                        "**hshstorefail" );
                    exit(-1);
                }
                FGP_DBG_PRINTF("Inserting in contextLeader_hshtbl, context-id=%d, LeaderWorldRank=%d, newcomm_ptr->co_shared_vars->rtw_map=%p\n", newcomm_ptr->context_id, LeaderWorldRank, newcomm_ptr->co_shared_vars->rtw_map);
            }

#else
        /* Notify the device of this new communicator */
	MPID_Dev_comm_create_hook( newcomm_ptr );
#endif
        
	*newcomm = newcomm_ptr->handle;
    }
    else {
	/* color was MPI_UNDEFINED.  Free the context id */
	*newcomm = MPI_COMM_NULL;
	MPIR_Free_contextid( new_context_id );
    }
    
    /* ... end of body of routine ... */

  fn_exit:
    MPIU_CHKLMEM_FREEALL();
    MPID_MPI_FUNC_EXIT(MPID_STATE_MPI_COMM_SPLIT);
    MPIU_THREAD_SINGLE_CS_EXIT("comm");
    return mpi_errno;
    
  fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
	    "**mpi_comm_split",
	    "**mpi_comm_split %C %d %d %p", comm, color, key, newcomm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}


#endif /* HK: Closing for #else USE_LID_LBI_CONTEXT */
